
#include <inttypes.h>
#include <xmmintrin.h>

/** \brief Merge s1 with s2, using s1 alpha channel, and save the result in d.
 *
 * This function merges two images using one of the images alpha channel.
 * The computation goes like this:
 *
 * *d = (s1.color * s1.alpha + s2.color * (255 - s1.alpha)) / 255;
 *
 * d can be set equal to s1 or s2 in which case that image is overwritten
 * with the result (in general d is set to s2.)
 *
 * All the buffers must be 16 bytes aligned and cannot be NULL pointers.
 *
 * The width must be a multiple of 4 RGBA pixels.
 *
 * Intrinsic headers are under:
 *
 * /usr/lib/x86_64-linux-gnu/gcc/x86_64-linux-gnu/\<gnu-version>/include/
 *
 * Use the following header to define them:
 *
 * #include <xmmintrin.h>
 *
 * There are many of them and they heavily depend on your architecture.
 *
 * My cpuinfo gives me the following list of flags on my i5 processor:
 *
 * flags: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat
 *        pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx
 *        rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology
 *        nonstop_tsc aperfmperf pni dtes64 monitor ds_cpl vmx smx est tm2
 *        ssse3 cx16 xtpr pdcm sse4_1 sse4_2 popcnt lahf_lm ida dts tpr_shadow
 *        vnmi flexpriority ept vpid
 */
void _merge_with_alpha_c(unsigned char *d, const unsigned char *s1, const unsigned char *s2, int w, int h)
{
	// the data is expected to be contiguous
	for(int i(w * h); i > 0; --i)
	{
		// at this point we do the same math as with the SSE4.2 implementation below
		// (i.e. avoid the r / 255, instead get a shift of 16)
		d[0] = ((s1[0] * s1[3] + s2[0] * (255 - s1[3])) * 257) / 65536;
		d[1] = ((s1[1] * s1[3] + s2[1] * (255 - s1[3])) * 257) / 65536;
		d[2] = ((s1[2] * s1[3] + s2[2] * (255 - s1[3])) * 257) / 65536;
		d[3] = 255;
		d += 4;
		s1 += 4;
		s2 += 4;
	}
}


void _merge_with_alpha_mmx(unsigned char *d, const unsigned char *s1, const unsigned char *s2, int w, int h)
{
	// these did not work as is, although it uses the stack when we could
	// just use 'offset[rip]' (see below) which is better optimized
	//const uint32_t explode_alpha_data[] = { 0x0C0C0C80, 0x08080880, 0x04040480, 0x00000080 };
	//const uint32_t explode_rgbo_data[] = { 0x800F800D, 0x800B8009, 0x80078005, 0x80038001 };
	//const uint32_t explode_rgbe_data[] = { 0x800E800C, 0x800A8008, 0x80068004, 0x80028000 };
	//const uint32_t m255_data[] = { 0xFFFFFF00, 0xFFFFFF00, 0xFFFFFF00, 0xFFFFFF00 };
	//const uint32_t m257_data[] = { 0x01010101, 0x01010101, 0x01010101, 0x01010101 };
	//const __m128 * const explode_alpha((__m128 const * const) explode_alpha_data);
	//const __m128 * const explode_rgbo((__m128 const * const) explode_rgbo_data);
	//const __m128 * const explode_rgbe((__m128 const * const) explode_rgbe_data);
	//const __m128 * const m255((__m128 const * const) m255_data);
	//const __m128 * const m257((__m128 const * const) m257_data);

	asm(
	// the data is expected to be contiguous
	"imul %[w],%[h]\n\t"
	"sar $2,%[h]\n\t"  // we do 4 pixels at a time, so divide by 4
	"jmp loop\n\t"
	".align 16\n"
"explode_alpha:\n\t"
	".long 0x80038003, 0x80078007, 0x800B800B, 0x800F800F\n"
"explode_rgbe:\n\t"
	".long 0x80028000, 0x80068004, 0x800A8008, 0x800E800C\n"
"explode_rgbo:\n\t"
	".long 0x80038001, 0x80078005, 0x800B8009, 0x800F800D\n"
"m255:\n\t"
	".long 0x00FF00FF, 0x00FF00FF, 0x00FF00FF, 0x00FF00FF\n"
"m257:\n\t"
	".long 0x01010101, 0x01010101, 0x01010101, 0x01010101\n"
"solid:\n\t"
	".long 0xFF000000, 0xFF000000, 0xFF000000, 0xFF000000\n"
"loop:\n\t"
	// load 4 pixels from both sources
	// (rrggbbaa:rrggbbaa:rrggbbaa:rrggbbaa)
	"movups (%[s1]),%%xmm1\n\t"
	"movups (%[s2]),%%xmm2\n\t"
	// extract the 4 alpha from source 1 (ignore source 2 alpha)
	// (00aa00aa:00aa00aa:00aa00aa:00aa00aa)
	"movdqa %%xmm1,%%xmm3\n\t"
	"pshufb explode_alpha(%%rip),%%xmm3\n\t"
	// compute (inverted = (255 - alpha <=> 255 ^ alpha))
	// (00ii00ii:00ii00ii:00ii00ii:00ii00ii)
	"movdqa %%xmm3,%%xmm4\n\t"
	"pxor m255(%%rip),%%xmm4\n\t"
	"movdqa %%xmm1,%%xmm5\n\t"
	"movdqa %%xmm2,%%xmm6\n\t"
	// explode two pixels per register (xmm1,2,5,6)
	// (even components: 00bb00rr:00bb00rr:00bb00rr:00bb00rr)
	// and multiply with alpha and (255 - alpha)
	"pshufb explode_rgbe(%%rip),%%xmm1\n\t"
	"pmullw %%xmm3,%%xmm1\n\t"
	"pshufb explode_rgbe(%%rip),%%xmm2\n\t"
	"pmullw %%xmm4,%%xmm2\n\t"
	// (odd components: 00aa00gg:00aa00gg:00aa00gg:00aa00gg)
	// and multiply with alpha and (255 - alpha)
	"pshufb explode_rgbo(%%rip),%%xmm5\n\t"
	"pmullw %%xmm3,%%xmm5\n\t"
	// even: sum both results (color * alpha) + (color * (255 - alpha))
	"paddw %%xmm1,%%xmm2\n\t"
	// (nearly) divide that result by 255, xmm2 is final: 00gg00aa:...
	"pmulhuw m257(%%rip),%%xmm2\n\t" // nearly equivalent to div m255(%rip),xmm2
	"pshufb explode_rgbo(%%rip),%%xmm6\n\t"
	"pmullw %%xmm4,%%xmm6\n\t"
	// odd: sum both results (color * alpha) + (color * (255 - alpha))
	"paddw %%xmm5,%%xmm6\n\t"
	// (nearly) divide that result by 255
	"pmulhuw m257(%%rip),%%xmm6\n\t" // nearly equivalent to div m255(%rip),xmm2
	// adjust the odd number back to the right place (colors << 8)
	// xmm6 is final: rr00bb00:...
	"psllw $8,%%xmm6\n\t"
	// combine with the color colors (rrggbbaa:...)
	"por %%xmm6,%%xmm2\n\t"
	"por solid(%%rip),%%xmm2\n\t"
	"movups %%xmm2,(%[d])\n\t"
	// move to the next pixel
	"add $16,%[s1]\n\t"
	"add $16,%[s2]\n\t"
	"add $16,%[d]\n\t"
	"dec %[h]\n\t" // [h] was set to ([w] * [h]) / 4 before the loop
	"jne loop\n\t"
	: /* output -- none */
	: /* input */ [d] "r" (d), [s1] "r" (s1), [s2] "r" (s2), [h] "r" (h), [w] "r" (w)
	: /*"%rbx", "%rcx", "%rsi", "%rdi",*/ "cc"
	);
}

// vim: ts=4 sw=4
